//
// Simple wrapper for OpenGL textures with different formats
//

#ifndef	__GL_TEXTURE__
#define	__GL_TEXTURE__

#ifdef  _WIN32
    #include    <windows.h>
	#pragma warning (disable:4996)
#endif

#include <GL/gl.h>
#include "libExt.h"

class	GlTexture
{
protected:

	GLuint	id;
	GLenum	target;
	GLenum  format, intFormat;
	int		width, height;
	GLenum	minFilter, magFilter;
	GLenum	wrap [3];					// only 3 coords can have wrap mode
	int		anisotropy;
	bool	bound;						// is texture bound now
	int		dirtyFlags;					// which attributes showld be reset on bind
	
public:

	GlTexture  ( int theWidth, int theHeight, GLenum theFormat, GLenum theIntFormat );
	~GlTexture ();

	bool	create       ();
	bool	bind         ();
	bool	unbind       ();
	bool	buildMipmaps ();
	
	bool	setMinFilter  ( GLenum filter = GL_LINEAR_MIPMAP_LINEAR );
	bool	setMagFilter  ( GLenum filter = GL_LINEAR );
	bool	setWrap       ( int s, int t, int r = GL_TEXTURE_WRAP_R );
	bool	setAnisotropy ( int aniso );
	
	bool	isOk () const
	{
		return target != 0 && id != 0 && width > 0 && height > 0;
	}
	
	int	getWidth () const
	{
		return width;
	}
	
	int	getHeight () const
	{
		return height;
	}
	
	GLuint	getId () const
	{
		return id;
	}
	
	GLenum	getFormat () const
	{
		return format;
	}
	
	GLenum	getInternalFormat () const
	{
		return intFormat;
	}
	
	GLenum	getMinFilter () const
	{
		return minFilter;
	}
	
	GLenum	getMagFilter () const
	{
		return magFilter;
	}
	
	int	getAnisotropy () const
	{
		return anisotropy;
	}
	
	GLenum	getWrap ( int coord ) const
	{
		return wrap [coord];
	}
	
	static	int		maxSize     ();
	
protected:
	enum						// dirty flags
	{
		minFilterFlag = 1,
		magFilterFlag = 2,
		wrapFlag      = 4,
		anisoFlag     = 8
	};
};

#endif
