//
//
//

#define	TEX_SIZE	512
#define	k1			2.345431f
#define	k2			1.12312312f
#define	k3			3.784728394f
#define	scale		0.03f

__global__ void animateKernel ( float4 * pos, float4 * norm, int w, int wh, int hh, float time )
{
	int		ix = blockIdx.x * blockDim.x + threadIdx.x;
	int		iy = blockIdx.y * blockDim.y + threadIdx.y;
	int		i  = iy * w + ix;
	float	x  = (ix - wh) * 0.03f;
	float	y  = (iy - hh) * 0.03f;
	float	v1 = x * 0.47f + y * 1.3f  + k1 * time;
	float	v2 = x * 2.9f  + y * 3.1f  + k2 * time;
	float	v3 = x * 5.577f  - y * 7.57f + k3 * time;
	float	nx = scale * ( 2.77f * 0.47f * __cosf ( v1 ) + 2.9f * __cosf ( v2 ) + 0.357f * 5.577f * __cosf ( v3 ) );
	float	ny = scale * ( 2.77f * 1.3f  * __cosf ( v1 ) + 3.1f * __cosf ( v2 ) - 0.357f * 7.57f * __cosf ( v3 ) );
	float	nz = 1.0f / sqrtf ( nx * nx + ny * ny );
	
	pos  [i] = make_float4 ( x, y, scale * ( 2.77f * __sinf ( v1 ) + __sinf ( v2 ) + 0.357f * __sinf ( v3 ) ), 1.0f );
	norm [i] = make_float4 ( nx, ny, nz, 0.0f );
}


extern "C" void buildVertices ( float4 * v, float4 * n, int w, int h, float time )
{
	dim3 threads ( 16, 16 );
	dim3 blocks  ( w / 16, h / 16 );
	
	animateKernel <<<blocks, threads>>> ( v, n, w, w / 2, h / 2, time );
	
    cudaThreadSynchronize ();
}
