//
// Example of image postprocessing via CUDA
// Computing is done from texture to VBO
// Image is copied to texture via ARB_pixel_buffer_object extensions
//
// Author: Alex V. Boreskoff <steps3d@gmail.com>, <steps3d@narod.ru>
//

#include    "libExt.h"

#ifdef	MACOSX
	#include	<GLUT/glut.h>
#else
	#include	<glut.h>
#endif

#include    <stdio.h>
#include    <stdlib.h>

#include    "libTexture.h"
#include    "TypeDefs.h"
#include    "Vector3D.h"
#include    "Vector2D.h"
#include    "boxes.h"
#include    "FrameBuffer.h"
#include    "GlslProgram.h"
#include	"CudaGlResource.h"

Vector3D    eye   ( -0.5, -0.5, 1.5 );  // camera position
unsigned    decalMap;                   // decal (diffuse) texture
unsigned    stoneMap;
unsigned    teapotMap;
unsigned    screenMap;
unsigned    outMap;
float       angle     = 0;
float       rot       = 0;
bool        useFilter = true;

CudaGlImage	 * src = NULL;
CudaGlBuffer * out = NULL;
VertexBuffer * imageBuffer;		// vertex coordinates
FrameBuffer	buffer ( 640, 480, FrameBuffer :: depth32 );
GlslProgram program1;

void    renderToBuffer ();
void 	postProcess    ();

extern "C" void doPostprocess ( cudaArray * inArray, uchar4 * outPtr, int w, int h );

void	createVertexBuffer ()
{
	imageBuffer = new VertexBuffer ();
	
	imageBuffer -> bind    ( GL_PIXEL_PACK_BUFFER_ARB );
	imageBuffer -> setData ( buffer.getWidth () * buffer.getHeight () * 4, NULL, GL_DYNAMIC_DRAW );
	imageBuffer -> unbind  ();
}

void init ()
{
    glClearColor ( 0.0, 0.0, 0.0, 1.0 );
    glEnable     ( GL_DEPTH_TEST );
    glEnable     ( GL_TEXTURE_2D );
    glDepthFunc  ( GL_LEQUAL     );

    glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
    glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void displayBoxes ()
{
    glMatrixMode ( GL_MODELVIEW );
    glPushMatrix ();

    glRotatef    ( rot, 0, 0, 1 );

    drawBox  ( Vector3D ( -5, -5, 0 ), Vector3D ( 10, 10, 3 ), stoneMap, false );
    drawBox  ( Vector3D ( 3, 2, 0.5 ), Vector3D ( 1,  2,  2 ), decalMap );

    glBindTexture   ( GL_TEXTURE_2D, teapotMap );
    glTranslatef    ( 0.2, 1, 1.5 );
    glRotatef       ( angle * 45.3, 1, 0, 0 );
    glRotatef       ( angle * 57.2, 0, 1, 0 );
    glutSolidTeapot ( 0.3 );

    glPopMatrix     ();
}

void    startOrtho ()
{
    glMatrixMode   ( GL_PROJECTION );                   // select the projection matrix
    glPushMatrix   ();                                  // store the projection matrix
    glLoadIdentity ();                                  // reset the projection matrix
                                                        // set up an ortho screen
    glOrtho        ( 0, buffer.getWidth (), 0, buffer.getHeight (), -1, 1 );
    glMatrixMode   ( GL_MODELVIEW );                    // select the modelview matrix
    glPushMatrix   ();                                  // store the modelview matrix
    glLoadIdentity ();                                  // reset the modelview matrix
}

void    endOrtho ()
{
    glMatrixMode ( GL_PROJECTION );                     // select the projection matrix
    glPopMatrix  ();                                    // restore the old projection matrix
    glMatrixMode ( GL_MODELVIEW );                      // select the modelview matrix
    glPopMatrix  ();                                    // restore the old projection matrix
}

void display ()
{
    renderToBuffer ();
	postProcess    ();	

    glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    startOrtho ();

	glEnable      ( GL_TEXTURE_RECTANGLE_ARB );
    glBindTexture ( GL_TEXTURE_RECTANGLE_ARB, outMap );

    glBegin ( GL_QUADS );

        glTexCoord2f ( 0, 0 );
        glVertex2f   ( 0, 0 );

        glTexCoord2f ( buffer.getWidth (), 0 );
        glVertex2f   ( buffer.getWidth (), 0 );

        glTexCoord2f ( buffer.getWidth (), buffer.getHeight () );
        glVertex2f   ( buffer.getWidth (), buffer.getHeight () );

        glTexCoord2f ( 0, buffer.getHeight () );
        glVertex2f   ( 0, buffer.getHeight () );
    glEnd   ();

	glDisable     ( GL_TEXTURE_RECTANGLE_ARB );
    glBindTexture ( GL_TEXTURE_RECTANGLE_ARB, 0 );
		
    endOrtho ();

    glutSwapBuffers ();
}

void reshape ( int w, int h )		
{
   glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   glMatrixMode   ( GL_PROJECTION );
   glLoadIdentity ();
   gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1.0, 60.0 );
   glMatrixMode   ( GL_MODELVIEW );
   glLoadIdentity ();
   gluLookAt      ( eye.x, eye.y, eye.z,    // eye
                    3, 3, 1,                // center
                    0, 0, 1 );              // up
}

void key ( unsigned char key, int x, int y )
{
    if ( key == 27 || key == 'q' || key == 'Q' )        // quit requested
        exit ( 0 );

    if ( key == 'f' || key == 'F' )
        useFilter = !useFilter;
}

void    specialKey ( int key, int x, int y )
{
    if ( key == GLUT_KEY_RIGHT )
        rot += 5;
    else
    if ( key == GLUT_KEY_LEFT )
        rot -= 5;

    glutPostRedisplay ();
}

void    animate ()
{
    angle  = 0.004f * glutGet ( GLUT_ELAPSED_TIME );
	
    glutPostRedisplay ();
}

void    renderToBuffer ()
{
	glBindTexture ( GL_TEXTURE_2D, 0 );

	program1.bind ();
	buffer.bind ();

    glClearColor ( 0, 0, 0, 1 );
    glClear      ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    reshape ( buffer.getWidth (), buffer.getHeight () );

    displayBoxes ();

    buffer.unbind ();
	program1.unbind ();
}

void postProcess ()
{
	src -> mapResource ();
	out -> mapResource ();

	size_t	    size;
	cudaArray * inArray  = src -> mappedArray ();
	uchar4    * res      = (uchar4 *)out -> mappedPointer ( size );
	
	doPostprocess ( inArray, res, buffer.getWidth (), buffer.getHeight () );
	
	src -> unmapResource ();
	out -> unmapResource ();
	
	imageBuffer -> bind ( GL_PIXEL_UNPACK_BUFFER_ARB );
	
	glBindTexture   ( GL_TEXTURE_RECTANGLE_ARB, outMap );
	glTexSubImage2D ( GL_TEXTURE_RECTANGLE_ARB, 0, 0, 0, buffer.getWidth (), buffer.getHeight (), GL_RGBA, GL_UNSIGNED_BYTE, NULL );
	
	imageBuffer -> unbind ();
	
	glBindBufferARB ( GL_PIXEL_UNPACK_BUFFER_ARB, 0 );
	glBindBufferARB ( GL_PIXEL_PACK_BUFFER_ARB, 0 );
}

GLuint createTextureDst ( int w, int h )
{
	GLuint	tex;
	
    glGenTextures ( 1, &tex );
    glBindTexture ( GL_TEXTURE_RECTANGLE_ARB, tex );

    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S,     GL_CLAMP_TO_EDGE );
    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T,     GL_CLAMP_TO_EDGE );
    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST       );
    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST       );
    glTexImage2D    ( GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA8, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL );
    glBindTexture   ( GL_TEXTURE_RECTANGLE_ARB, 0 );
	
	return tex;
}

int main ( int argc, char * argv [] )
{
                                // initialize glut
    glutInit            ( &argc, argv );
    glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH );
    glutInitWindowSize  ( buffer.getWidth (), buffer.getHeight () );

                                // create window
    glutCreateWindow ( "OpenGL/CUDA interoperability: postprocessing" );

                                // register handlers
    glutDisplayFunc  ( display    );
    glutReshapeFunc  ( reshape    );
    glutKeyboardFunc ( key        );
    glutSpecialFunc  ( specialKey );
    glutIdleFunc     ( animate    );

    init           ();
    initExtensions ();

    assertExtensionsSupported ( "GL_ARB_shading_language_100 GL_ARB_shader_objects EXT_framebuffer_object ARB_texture_rectangle" );

    decalMap  = createTexture2D ( true, "../../Textures/oak.bmp" );
    stoneMap  = createTexture2D ( true, "../../Textures/block.bmp" );
    teapotMap = createTexture2D ( true, "../../Textures/Oxidated.jpg" );
	outMap    = createTextureDst ( buffer.getWidth (), buffer.getHeight () );
	screenMap = buffer.createColorTexture ( GL_RGBA );

	buffer.create ();
	buffer.bind   ();
	buffer.attachColorTexture ( GL_TEXTURE_2D, screenMap );

	if ( !buffer.isOk () )
		printf ( "Error with framebuffer\n" );

	buffer.unbind ();

    if ( !program1.loadShaders ( "draw.vsh", "draw.fsh" ) )
    {
        printf ( "Error loading draw shaders:\n%s\n", program1.getLog ().c_str () );

        return 3;
    }

    program1.bind            ();
    program1.setTexture      ( "texImage", 0 );
    program1.unbind          ();

						// initialize CUDA device for OpenGL
    cudaGLSetGLDevice ( 0 );
	
	src = new CudaGlImage ( buffer.getColorBuffer (), GL_TEXTURE_2D, cudaGraphicsMapFlagsNone );
	
	createVertexBuffer ();

	out  = new CudaGlBuffer ( imageBuffer, GL_ARRAY_BUFFER );
	
    glutMainLoop ();

    return 0;
}
