//
// Example of image postprocessing via CUDA
//
// Author: Alex V. Boreskoff <steps3d@gmail.com>, <steps3d@narod.ru>
//

#include    "libExt.h"

#ifdef	MACOSX
	#include	<GLUT/glut.h>
#else
	#include	<glut.h>
#endif

#include    <stdio.h>
#include    <stdlib.h>

#include    "libTexture.h"
#include    "TypeDefs.h"
#include    "Vector3D.h"
#include    "Vector2D.h"
#include    "boxes.h"
#include    "FrameBuffer.h"
#include    "GlslProgram.h"

#include	"CudaGlResource.h"

Vector3D    eye   ( -0.5, -0.5, 1.5 );  // camera position
unsigned    decalMap;                   // decal (diffuse) texture
unsigned    stoneMap;
unsigned    teapotMap;
unsigned    screenMap;
unsigned    outMap;
float       angle     = 0;
float       rot       = 0;
bool        useFilter = true;

CudaGlImage	* src = NULL;
CudaGlImage * dst = NULL;
uchar4      * buf = NULL;
FrameBuffer	buffer ( 640, 480, FrameBuffer :: depth32 );
GlslProgram program1;					// render to uchar4 texture
GlslProgram program2;					// render to normalized color buffer

void    renderToBuffer ();
void 	postProcess    ();

extern "C" void doPostprocess ( cudaArray * inArray, cudaArray * outArray, uchar4 * buf, int pitch, int w, int h );

void init ()
{
    glClearColor ( 0.0, 0.0, 0.0, 1.0 );
    glEnable     ( GL_DEPTH_TEST );
    glEnable     ( GL_TEXTURE_2D );
    glDepthFunc  ( GL_LEQUAL     );

    glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
    glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void displayBoxes ()
{
    glMatrixMode ( GL_MODELVIEW );
    glPushMatrix ();

    glRotatef    ( rot, 0, 0, 1 );

    drawBox  ( Vector3D ( -5, -5, 0 ), Vector3D ( 10, 10, 3 ), stoneMap, false );
    drawBox  ( Vector3D ( 3, 2, 0.5 ), Vector3D ( 1,  2,  2 ), decalMap );

    glBindTexture   ( GL_TEXTURE_2D, teapotMap );
    glTranslatef    ( 0.2, 1, 1.5 );
    glRotatef       ( angle * 45.3, 1, 0, 0 );
    glRotatef       ( angle * 57.2, 0, 1, 0 );
    glutSolidTeapot ( 0.3 );

    glPopMatrix     ();
}

void    startOrtho ()
{
    glMatrixMode   ( GL_PROJECTION );                   // select the projection matrix
    glPushMatrix   ();                                  // store the projection matrix
    glLoadIdentity ();                                  // reset the projection matrix
                                                        // set up an ortho screen
    glOrtho        ( 0, buffer.getWidth (), 0, buffer.getHeight (), -1, 1 );
    glMatrixMode   ( GL_MODELVIEW );                    // select the modelview matrix
    glPushMatrix   ();                                  // store the modelview matrix
    glLoadIdentity ();                                  // reset the modelview matrix
}

void    endOrtho ()
{
    glMatrixMode ( GL_PROJECTION );                     // select the projection matrix
    glPopMatrix  ();                                    // restore the old projection matrix
    glMatrixMode ( GL_MODELVIEW );                      // select the modelview matrix
    glPopMatrix  ();                                    // restore the old projection matrix
}

void display ()
{
    renderToBuffer ();
	postProcess    ();
	
    glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    startOrtho ();

	program2.bind ();

	glEnable      ( GL_TEXTURE_2D );
    glBindTexture ( GL_TEXTURE_2D, outMap );

    glBegin ( GL_QUADS );

        glTexCoord2f ( 0, 0 );
        glVertex2f   ( 0, 0 );

        glTexCoord2f ( 1, 0 );
        glVertex2f   ( buffer.getWidth (), 0 );

        glTexCoord2f ( 1, 1 );
        glVertex2f   ( buffer.getWidth (), buffer.getHeight () );

        glTexCoord2f ( 0, 1 );
        glVertex2f   ( 0, buffer.getHeight () );
    glEnd   ();
	
	glDisable     ( GL_TEXTURE_2D );
    glBindTexture ( GL_TEXTURE_2D, 0 );
		
	program2.unbind ();
		
    endOrtho ();

    glutSwapBuffers ();
}

void reshape ( int w, int h )		
{
   glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   glMatrixMode   ( GL_PROJECTION );
   glLoadIdentity ();
   gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1.0, 60.0 );
   glMatrixMode   ( GL_MODELVIEW );
   glLoadIdentity ();
   gluLookAt      ( eye.x, eye.y, eye.z,    // eye
                    3, 3, 1,                // center
                    0, 0, 1 );              // up
}

void key ( unsigned char key, int x, int y )
{
    if ( key == 27 || key == 'q' || key == 'Q' )        // quit requested
        exit ( 0 );

    if ( key == 'f' || key == 'F' )
        useFilter = !useFilter;
}

void    specialKey ( int key, int x, int y )
{
    if ( key == GLUT_KEY_RIGHT )
        rot += 5;
    else
    if ( key == GLUT_KEY_LEFT )
        rot -= 5;

    glutPostRedisplay ();
}

void    animate ()
{
    angle  = 0.004f * glutGet ( GLUT_ELAPSED_TIME );
	
    glutPostRedisplay ();
}

void    renderToBuffer ()
{
	glBindTexture ( GL_TEXTURE_2D, 0 );

	program1.bind ();
	buffer.bind ();

    glClearColor ( 0, 0, 0, 1 );
    glClear      ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    reshape ( buffer.getWidth (), buffer.getHeight () );

    displayBoxes ();

    buffer.unbind ();
	program1.unbind ();
}

void postProcess ()
{
	src -> mapResource ();
	dst -> mapResource ();

	cudaArray * inArray  = src -> mappedArray ();
	cudaArray * outArray = dst -> mappedArray ();
	
	doPostprocess ( inArray, outArray, buf, 0, buffer.getWidth (), buffer.getHeight () );
	
	src -> unmapResource ();
	dst -> unmapResource ();	
}

GLuint createTextureDst ( int w, int h )
{
	GLuint	tex;
	
    glGenTextures ( 1, &tex );
    glBindTexture ( GL_TEXTURE_2D, tex );

    glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,     GL_CLAMP_TO_EDGE );
    glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,     GL_CLAMP_TO_EDGE );
    glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST       );
    glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST       );

    glTexImage2D ( GL_TEXTURE_2D, 0, GL_RGBA8UI_EXT, w, h, 0, GL_RGBA_INTEGER_EXT, GL_UNSIGNED_BYTE, NULL );
	
	return tex;
}

/*
// This function returns the best GPU (with maximum GFLOPS)
int getMaxGflopsDeviceId ()
{
	int current_device   = 0, sm_per_multiproc = 0;
	int max_compute_perf = 0, max_perf_device  = 0;
	int device_count     = 0, best_SM_arch     = 0;
    int arch_cores_sm [3] = { 1, 8, 32 };
	cudaDeviceProp deviceProp;

	cudaGetDeviceCount( &device_count );
	
	// Find the best major SM Architecture GPU device
	while ( current_device < device_count ) {
		cudaGetDeviceProperties( &deviceProp, current_device );
		if (deviceProp.major > 0 && deviceProp.major < 9999) {
			best_SM_arch = MAX(best_SM_arch, deviceProp.major);
		}
		current_device++;
	}

    // Find the best CUDA capable GPU device
	current_device = 0;
	while( current_device < device_count ) {
		cudaGetDeviceProperties( &deviceProp, current_device );
		if (deviceProp.major == 9999 && deviceProp.minor == 9999) {
		    sm_per_multiproc = 1;
		} else if (deviceProp.major <= 2) {
			sm_per_multiproc = arch_cores_sm[deviceProp.major];
		} else {
			sm_per_multiproc = arch_cores_sm[2];
		}

		int compute_perf  = deviceProp.multiProcessorCount * sm_per_multiproc * deviceProp.clockRate;
		if( compute_perf  > max_compute_perf ) {
            // If we find GPU with SM major > 2, search only these
			if ( best_SM_arch > 2 ) {
				// If our device==dest_SM_arch, choose this, or else pass
				if (deviceProp.major == best_SM_arch) {	
					max_compute_perf  = compute_perf;
					max_perf_device   = current_device;
				}
			} else {
				max_compute_perf  = compute_perf;
				max_perf_device   = current_device;
			}
		}
		++current_device;
	}
	return max_perf_device;
}
*/

int main ( int argc, char * argv [] )
{
                                // initialize glut
    glutInit            ( &argc, argv );
    glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH );
    glutInitWindowSize  ( buffer.getWidth (), buffer.getHeight () );

                                // create window
    glutCreateWindow ( "OpenGL/CUDA interoperability: postprocessing" );

                                // register handlers
    glutDisplayFunc  ( display    );
    glutReshapeFunc  ( reshape    );
    glutKeyboardFunc ( key        );
    glutSpecialFunc  ( specialKey );
    glutIdleFunc     ( animate    );

    init           ();
    initExtensions ();

    assertExtensionsSupported ( "GL_ARB_shading_language_100 GL_ARB_shader_objects EXT_framebuffer_object ARB_texture_rectangle" );

	outMap    = createTextureDst ( buffer.getWidth (), buffer.getHeight () );
    decalMap  = createTexture2D ( true, "../../Textures/oak.bmp" );
    stoneMap  = createTexture2D ( true, "../../Textures/block.bmp" );
    teapotMap = createTexture2D ( true, "../../Textures/Oxidated.jpg" );
	screenMap = buffer.createColorTexture ( GL_RGBA_INTEGER_EXT, GL_RGBA8UI_EXT );

	buffer.create ();
	buffer.bind   ();
	buffer.attachColorTexture ( GL_TEXTURE_2D, screenMap );

	if ( !buffer.isOk () )
		printf ( "Error with framebuffer\n" );

	buffer.unbind ();

    if ( !program1.loadShaders ( "draw.vsh", "draw-uint.fsh" ) )
    {
        printf ( "Error loading draw-uint shaders:\n%s\n", program1.getLog ().c_str () );

        return 3;
    }
	
    program1.bind            ();
    program1.setTexture      ( "texImage", 0 );
    program1.unbind          ();

    if ( !program2.loadShaders ( "draw.vsh", "draw-color.fsh" ) )
    {
        printf ( "Error loading draw-uint shaders:\n%s\n", program2.getLog ().c_str () );

        return 3;
    }
	
    program2.bind            ();
    program2.setTexture      ( "texImage", 0 );
    program2.unbind          ();
	
						// initialize CUDA device for OpenGL
    cudaGLSetGLDevice ( 0 );
	cudaMalloc        ( (void **) &buf, buffer.getWidth () * buffer.getHeight () * 4 );
	
	src = new CudaGlImage ( buffer.getColorBuffer (), GL_TEXTURE_2D, cudaGraphicsMapFlagsNone );
	dst = new CudaGlImage ( outMap, GL_TEXTURE_2D, cudaGraphicsMapFlagsNone );	
	
    glutMainLoop ();

    return 0;
}
