//
// Example of image postprocessing via CUDA
//
// Author: Alex V. Boreskoff <steps3d@gmail.com>, <steps3d@narod.ru>
//

#include    "libExt.h"

#ifdef	MACOSX
	#include	<GLUT/glut.h>
#else
	#include	<glut.h>
#endif

#include    <stdio.h>
#include    <stdlib.h>

#include    "libTexture.h"
#include    "TypeDefs.h"
#include    "Vector3D.h"
#include    "Vector2D.h"
#include    "boxes.h"
#include    "GlslProgram.h"
#include	"CudaGlResource.h"

#define	TEX_SIZE	512				// width & height of texture

Vector3D    eye   ( -10, -10, 5 );  // camera position
Vector3D	light ( 5, 0, 4 );		// light position
float	 	angle = 0;
Vector3D	rot ( 0, 0, 0 );
int			mouseOldX = 0;
int			mouseOldY = 0;
float		waveScale = 0.2;
float		time      = 0.0f;

unsigned	reflectionMap;			// cubic map with reflection

CudaGlBuffer * pos  = NULL;
CudaGlBuffer * norm = NULL;
VertexBuffer * vertexBuffer;		// vertex coordinates
VertexBuffer * normalBuffer;		// normal buffer
VertexBuffer * indexBuffer;			// vertex index buffer
GlslProgram    program;

extern "C" void buildVertices ( float4 * v, float4 * n, int w, int h, float time );

void	createIndexBuffer ()
{
	int * data = (int *) malloc ( (TEX_SIZE-1)*(TEX_SIZE-1)*6*sizeof (int) );
	int	  k    = 0;
	
	for ( int i = 0; i < TEX_SIZE - 1; i++ )
		for ( int j = 0; j < TEX_SIZE - 1; j++ )
		{
			data [k]   = i + TEX_SIZE*j;				// first triangle (i,j)-(i+1,j) - (i,j+1)
			data [k+1] = i + TEX_SIZE*j + 1;
			data [k+2] = i + TEX_SIZE*j + TEX_SIZE;
			
			data [k+3] = i + TEX_SIZE*j + 1;			// second triangle (i+1,j)-(i+1,j+1) - (i,j+1)
			data [k+4] = i + TEX_SIZE*j + TEX_SIZE + 1;
			data [k+5] = i + TEX_SIZE*j + TEX_SIZE;
			
			k         += 6;
		}
		
	indexBuffer = new VertexBuffer ();
	
	indexBuffer -> bind    ( GL_ELEMENT_ARRAY_BUFFER_ARB );
	indexBuffer -> setData ( (TEX_SIZE - 1) * (TEX_SIZE - 1) * 6 * sizeof ( int ), data, GL_STATIC_DRAW );
	indexBuffer -> unbind  ();
	
	free ( data );
}

void	createVertexBuffer ()
{
	vertexBuffer = new VertexBuffer ();
	
	vertexBuffer -> bind    ( GL_PIXEL_PACK_BUFFER_ARB );
	vertexBuffer -> setData ( TEX_SIZE * TEX_SIZE * 4 * sizeof ( float ), NULL, GL_DYNAMIC_DRAW );
	vertexBuffer -> unbind  ();
}

void	createNormalBuffer ()
{
	normalBuffer = new VertexBuffer ();
	
	normalBuffer -> bind    ( GL_PIXEL_PACK_BUFFER_ARB );
	normalBuffer -> setData ( TEX_SIZE * TEX_SIZE * 4 * sizeof ( float ), NULL, GL_DYNAMIC_DRAW );
	normalBuffer -> unbind  ();
}

void init ()
{
    glClearColor ( 0.0, 0.0, 0.0, 1.0 );
    glEnable     ( GL_DEPTH_TEST );
    glEnable     ( GL_TEXTURE_2D );
    glDepthFunc  ( GL_LEQUAL     );

    glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
    glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void display ()
{
	float	scale = 1.0f;
	size_t	vSize, nSize;
	
	pos  -> mapResource ();
	norm -> mapResource ();
	
	buildVertices ( (float4 *) pos  -> mappedPointer ( vSize ), 
					(float4 *) norm -> mappedPointer ( nSize ), TEX_SIZE, TEX_SIZE, time );
	
	pos  -> unmapResource ();
	norm -> unmapResource ();
	
	glClear      ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glMatrixMode ( GL_MODELVIEW );
	glPushMatrix ();

    glRotatef    ( rot.x, 1, 0, 0 );
    glRotatef    ( rot.y, 0, 1, 0 );
    glRotatef    ( rot.z, 0, 0, 1 );
	glScalef     ( scale, scale, 2 );

	glActiveTextureARB   ( GL_TEXTURE0_ARB );
    glBindTexture        ( GL_TEXTURE_CUBE_MAP, reflectionMap );
	
    glPushClientAttrib   ( GL_CLIENT_VERTEX_ARRAY_BIT );
	
	glEnableClientState  ( GL_VERTEX_ARRAY );
	vertexBuffer -> bind ( GL_ARRAY_BUFFER_ARB );
	glVertexPointer      ( 4, GL_FLOAT, 0, NULL );
	
	glEnableClientState  ( GL_TEXTURE_COORD_ARRAY );
	normalBuffer -> bind ( GL_ARRAY_BUFFER_ARB );
	glTexCoordPointer    ( 4, GL_FLOAT, 0, NULL );
	
    glEnableClientState  ( GL_INDEX_ARRAY  );
	indexBuffer  -> bind ( GL_ELEMENT_ARRAY_BUFFER_ARB );
    glIndexPointer       ( GL_UNSIGNED_INT, 0, 0 );
	
	program.bind   ();

	glDrawElements     ( GL_TRIANGLES, (TEX_SIZE-1)*(TEX_SIZE-1)*6, GL_UNSIGNED_INT, 0 );
	program.unbind ();

	glPopClientAttrib ();
	glPopMatrix       ();
	glutSwapBuffers   ();
}

void reshape ( int w, int h )		
{
	glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
	glMatrixMode   ( GL_PROJECTION );
	glLoadIdentity ();
	gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1, 5000.0 );
	gluLookAt      ( eye.x, eye.y, eye.z,	// eye
					 0, 0, 0,				// center
					 0, 0, 1 );				// up


	glMatrixMode   ( GL_MODELVIEW );
	glLoadIdentity ();
}

void motion ( int x, int y )
{
    rot.y -= ((mouseOldY - y) * 180.0f) / 200.0f;
    rot.z -= ((mouseOldX - x) * 180.0f) / 200.0f;
    rot.x  = 0;

    if ( rot.z > 360 )
		rot.z -= 360;

	if ( rot.z < -360 )
		rot.z += 360;

    if ( rot.y > 360 )
		rot.y -= 360;

	if ( rot.y < -360 )
        rot.y += 360;

    mouseOldX = x;
    mouseOldY = y;

	glutPostRedisplay ();
}

void mouse ( int button, int state, int x, int y )
{
    if ( state == GLUT_DOWN )
    {
        mouseOldX = x;
        mouseOldY = y;
	}
}

void key ( unsigned char key, int x, int y )
{
	if ( key == 27 || key == 'q' || key == 'Q' )	//	quit requested
    	exit ( 0 );
		
	if ( key == '+' )
		waveScale += 0.01;
	else
	if ( key == '-' )
		waveScale -= 0.01;
}

void    animate ()
{
    time = 0.001f * glutGet ( GLUT_ELAPSED_TIME );
	
    glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
                                // initialize glut
    glutInit            ( &argc, argv );
    glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH );
    glutInitWindowSize  ( 640, 480 );


                                // create window
    glutCreateWindow ( "OpenGL/CUDA interoperability: vertex buffers" );

                                // register handlers
	glutDisplayFunc  ( display );
	glutReshapeFunc  ( reshape );
	glutKeyboardFunc ( key     );
	glutMouseFunc    ( mouse   );
	glutMotionFunc   ( motion  );
	glutIdleFunc     ( animate );

    init           ();
    initExtensions ();

	if ( !GlslProgram :: isSupported () )
	{
		printf ( "GLSL not supported\n" );
		
		return 1;
	}

	const char * faces [6] =
	{
		"../../Textures/Cubemaps/skyrt.bmp",
		"../../Textures/Cubemaps/skylf.bmp",
		"../../Textures/Cubemaps/skydn.bmp",
		"../../Textures/Cubemaps/skyup.bmp",
		"../../Textures/Cubemaps/skybk.bmp",
		"../../Textures/Cubemaps/skyfr.bmp",
	};

    if ( !program.loadShaders ( "water-draw.vsh", "water-draw.fsh" ) )
    {
        printf ( "Error loading water-draw shaders:\n%s\n", program.getLog ().c_str () );

        return 3;
    }
	
    program.bind             ();
    program.setTexture       ( "reflectionMap", 0 );
	program.setUniformVector ( "eye",  eye  );
    program.unbind           ();

	reflectionMap = createCubeMap   ( true, faces );

						// initialize CUDA device for OpenGL
    cudaGLSetGLDevice  ( 0 );
	createVertexBuffer ();
	createNormalBuffer ();
	createIndexBuffer  ();

	pos  = new CudaGlBuffer ( vertexBuffer, GL_ARRAY_BUFFER );
	norm = new CudaGlBuffer ( normalBuffer, GL_ARRAY_BUFFER );
	
    glutMainLoop ();

    return 0;
}
